import sys
import os
import re
from datetime import datetime

def simpleName(inputPath):
   """Return the last element of a path to display just the input name"""
   justInput=re.search('([^/]+)/?$', inputPath)   
   return justInput.group(0)  

def logStart(inputList, dataList):
   """Creates the log file 'ComplexConstructorLog' and writes initial message in log file
      with the date and the input files"""
   now = datetime.now() 
   dt_string = now.strftime("%d/%m/%Y %H:%M:%S")
   sys.stderr = open("./"+dataList[2]+"/ComplexConstructor.log", "w")
   sys.stderr.write("\n                              ComplexConstructor\n")
   sys.stderr.write("     BUILDING OF COMPLEXES USING THE PAIRED INTERACTION OF ITS ELEMENTS\n")
   sys.stderr.write("Written by Paula Gomis Rosa, Arturo González Vilanova and Marta López Balastegui\n\n")
   print("Job starting time: ", dt_string, file=sys.stderr)
   sys.stderr.write("\nInput files:\n\t - FASTA sequences in file:\n\t ")   
   fas=inputList[-1]
   print('\t',simpleName(fas), file=sys.stderr)
   sys.stderr.write("\n\t - PDB files:\n")  
   pdbs=list( inputList[i] for i in range(len(inputList)-1) )
   for element in pdbs:
      print('\t\t',simpleName(element), file=sys.stderr)
   return None

def progress(outDir,dic,stValues):
   """Write in the log file the progression information"""
   sys.stderr = open("./"+outDir+"/ComplexConstructor.log", "a")

   keyList=[]
   for element in dic:
      keyList.append(element)

   print("\nNumber of chains found: ", len(keyList), file=sys.stderr) 
   print("IDs of the chains: ", (', '.join(keyList)), file=sys.stderr)

   if stValues is not None:
      print("\nThe stoichiometry for those chains is: ", file=sys.stderr)
      for element in stValues:
         print(element,": ",stValues[element], file=sys.stderr)
   print("\n", file=sys.stderr)
   return None

def end(outDir, outFile):
   """ Final message in logFile """
   cwd = os.getcwd()
   resultDir=cwd+'/'+outDir
   print("\nProcess completed correctly, macrocomplex structure in file '%s'" %outFile,file=sys.stderr)
   print("\nThe result is stored in directory '%s'." %resultDir, file=sys.stderr)
   return None


from tkinter import filedialog
from tkinter import messagebox
from tkinter import *
import os
import re
import complexconstructor.helpText as helpText
import webbrowser

def initGui():
    cwd = os.getcwd()
    #initialize main window
    root = Tk()
    logo = Image("photo", file = os.path.dirname(os.path.realpath(__file__))+"/CB_icon.png")
    logo_sm = Image("photo", file = os.path.dirname(os.path.realpath(__file__))+"/CB_icon_sm.png")
    root.title("ComplexConstructor")
    root.call('wm','iconphoto',root._w, logo)
    root.configure(background = "#2da6e6")

    def simpleName(inputPath):
        """Returns the last element of a path to display just the input name"""
        justInput=re.search('([^/]+)/?$', inputPath)
        return justInput.group(0)

    def fastaSelection():
        """Opens window with *.fa files in the current directory to select the FASTA file"""
        filetext="Select"
        root.filename =  filedialog.askopenfilename(initialdir = cwd,title = "Select the FASTA file",
                                                    filetypes = (("FASTA files","*.fa"),
                                                                ("all files","*.*")))
        print (root.filename)
        bF["text"]= simpleName(root.filename) if root.filename else filetext
        return root.filename

    def pdbSelection():
        """Opens window with directories present in the current folder to select the the one
        with the pair interactions in PDB"""
        filetext="Select"
        root.directory = filedialog.askdirectory(title = "Select the directory with PDB files")
        print (root.directory)
        bP["text"]= simpleName(root.directory) if root.directory else filetext
        return root.directory

    def oSelection():
        """Stores the name introduced for the output file """
        filetext="Confirm"
        print (outputName.get())
        bO["text"]= simpleName(outputName.get()) if outputName.get() else filetext
        return outputName.get()

    def stSelection():
        """Opens window with files in the current directory to select the the stoichiometry file"""
        filetext="Select"
        root.filenameSt =  filedialog.askopenfilename(initialdir = cwd,title = "Select the stoichiometry file",
                                                    filetypes = (("Text files","*.txt"),
                                                                ("all files","*.*")))
        print (root.filenameSt)
        frameSt.pack(pady="15", padx="50",anchor=W, fill=X, expand=YES)
        bSt["text"]= simpleName(root.filenameSt) if root.filenameSt else filetext
        labelO["text"]="Output directory name:                       "
        labelSt.lift()
        labelSt.pack(side=LEFT )
        bSt.lift()
        bSt.pack()
        return root.filenameSt

    def checkVerbose():
        """Returns TRUE if the log checkbox if checked"""
        return verb.get()

    def showHelp():
        """Displays help of the ComplexConstructor in a new window"""
        t = Toplevel(root)
        t.wm_title("Help")
        l = Label(t, text=helpText.helpMessage,justify=LEFT)
        l.pack(side="top", fill="both", expand=True, padx=100, pady=100)

    def openTutorial():
        """Displays README file from gitHub"""
        webbrowser.open_new("https://github.com/Argonvi/SBI-PYT_Project/blob/master/README.md")

    def runCB():
        """Checks that all the mandatory inputs are defined and run the program"""
        if outputName.get()!='':
            try:
                print(root.filename)
                print(root.directory )
                root.destroy()
            except:
                messagebox.showerror("Ups!", "You should select the input FASTA file and the directory with the PDB interaction files.")

        else:
            messagebox.showerror("Ups!", "You should specify and confirm the name of the directory where the output files will be stored.")


    # Setting of the topmenu
    menubar = Menu(root, tearoff=0)
    filemenu = Menu(menubar)

    filemenu.add_command(label="Add stoichiometry", command=stSelection)
    #filemenu.add_command(label="Energy analysis", command=menu_action)
    #filemenu.add_command(label="Open log file", command=menu_action)
    filemenu.add_separator()
    filemenu.add_command(label="Quit", command=root.quit)

    helpmenu = Menu(menubar)
    helpmenu.add_command(label="Help", command=showHelp)
    helpmenu.add_separator()
    helpmenu.add_command(label="About", command=openTutorial)

    menubar.add_cascade(label="Options", menu=filemenu)
    menubar.add_cascade(label="Help", menu=helpmenu)

    root.config(menu=menubar)


    #Console frame
    frame = Frame(root)
    frame.pack(pady="10", padx="10")

    frameWll = Frame(frame)
    frameWll.pack(pady="10")

    #Initial text
    panel = Label(frameWll, image = logo_sm)
    panel.pack(side=LEFT, padx="9")
    wellcome = Label(frameWll, text="Introduce the FASTA file and the directory with the interactions in PDB:    ")
    wellcome.pack(side=TOP,pady="10")


    # FASTA widget
    frameFA = Frame(frame)
    frameFA.pack(pady="15", padx="50",side=TOP,anchor=W, fill=X, expand=YES)
    labelF = Label(frameFA, text="Sequence file:                                          ")
    labelF.pack( side=LEFT )
    bF = Button(frameFA, text="Select", command=fastaSelection) #Button FASTA file
    bF.pack( )


    #PDB widget
    framePDB = Frame(frame)
    framePDB.pack(pady="15", padx="50",anchor=W, fill=X, expand=YES)
    labelF = Label(framePDB, text="Interactions directory:                             ")
    labelF.pack( side=LEFT )
    bP = Button(framePDB, text="Select", command=pdbSelection)  #Button PDB dir
    bP.pack( )


    #Output widget
    frameO = Frame(frame)
    frameO.pack(pady="15", padx="50",anchor=W, fill=X, expand=YES)
    labelO = Label(frameO, text="Output directory name:    ")
    labelO.pack( side=LEFT )
    outputName = StringVar()  #textBox for output folder
    nameEntered = Entry(frameO, width = 15, textvariable = outputName)
    nameEntered.pack(side=LEFT, padx="5")
    bO = Button(frameO, text = "Confirm", command = oSelection)
    bO.pack(side=LEFT)


    #Verbose widget
    frameV = Frame(frame)
    frameV.pack(pady="15", padx="50",side=LEFT,anchor=W, fill=X, expand=YES)
    labelF = Label(frameV, text="Do you want to create a log file?")
    labelF.pack( side=LEFT )
    verb = BooleanVar()  #Checkbox verbose
    cV = Checkbutton(frameV, text="I do!", variable=verb, command=checkVerbose)
    cV.pack()

    #Stoichiometry widget
    frameSt = Frame(frame)
    frameSt.pack_forget()
    labelSt = Label(frameSt, text="Define a stoichiometry:  ")
    labelSt.lower()
    bSt = Button(frameSt, text = "Select", command = stSelection)
    bSt.lower()


    #Submit widget
    bEnter = Button(root, text="Build complex", command=runCB, bd="2",
                     relief="solid",padx="10", pady="10") #fg="red", bg="blue",
    bEnter.pack( side=BOTTOM ,pady="5")
    root.mainloop()


    #Parse inputs
    inputs=[]
    inputs.append(root.filename) # FASTA
    inputs.append(root.directory) # PDB dir
    inputs.append(checkVerbose()) # verbose T/F
    try: # stoichiometry None/file
        inputs.append( root.filenameSt)
    except:
        inputs.append(None)
    inputs.append(outputName.get()) #Output folder
    print(inputs)
    return inputs



import sys,os,argparse, gzip
from os.path import isfile, join


def readArgs():
    """ Read and organize the command line arguments and return the namespace"""
    parser = argparse.ArgumentParser(description="Build a protein complex from a set "
                                                 "of pdb files containing the paired "
                                                 "structures of its elements.")

    parser.add_argument('-fa', '--fasta', dest = "infasta", action = "store", default = None, 
                        help = """FASTA file with the sequences of the proteins
                               or DNA that will conform the complex.""")

    parser.add_argument('-pdb', '--pdbDir', dest = "inpdb", action = "store", default = None, 
                        help = """Diretory containing the PDB files with the 
                               structure of the pairs that will conform the complex.""")                        

    parser.add_argument('-o', '--output', dest = "outfile", action = "store", default = None, 
                        help = """Directory name where the complex results will be stored. 
                                """)

    parser.add_argument('-v', '--verbose', dest = "verbose", action = "store_true", default = False, 
                        help = """Show the detailed progression of the building process 
                                in a file called ComplexConstructor.log.""")

    parser.add_argument('-st', '--stoichiometry', dest = "stoich", action = "store", default = None, 
                        help = """File containing a determined stoichiometry to the complex. 
                               The information of the stoichiometry must be: the ID of the 
                               sequence chain (concordant with the FASTA file ID) followed by 
                               the number of times it has to be present in the complex after ':'
                               ID_as_FASTA_file : stoichiometry (one per line) in format .txt. """) 

    parser.add_argument('-gui', '--graphicInterface', dest="gui",action="store_true",default=False,
                        help="""To use ComplexConstructor with the graphical interface just use 
                              '-gui' argument in commandline.""") 

    return parser.parse_args()







helpMessage="""ComplexConstructor:

To build the macrocomplex fill in the main window requirements. As for running via 
command-line, a FASTA file with the sequences and a directory with the PDB files 
of the interacting are required. In addition, a name for the folder where the results 
will be stored is needed, after typing it you should confirm it. 

Note that, to select the PDB directory you have to enter in the desired directory 
and then select it. 

Furthermore, additional options can be set:

    In the main window you can specify if you want to create a log file where the 
    process of the execution will be displayed. 

    In the top menu, in the 'Options' dropdown there is the 'Add Stoichiometry' 
    option. You can upload a file with a determined stoichiometry to be applied 
    to the macrocomplex. The format of this file has to be the ID of the sequence, 
    concordant with the one in the FASTA file, followed by ':' and a number. This 
    number will be the number of times the corresponging sequence will be in the 
    final complex."""


import os
from os.path import isfile, join
import sys
from Bio.PDB import *
from Bio import SeqIO, pairwise2
import copy
import complexconstructor.logProgress as logProgress

alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"

class sequence_clashing_error(Exception):
    """Error due to clashes between new added chain and the previous structure"""
    def __init__(self, chain):
        self.chain=chain
    def __str__(self):
        return "The chain %s can't be added as it clashes with the complex." % (self.chain.get_id())

class different_length_error(Exception):
    """Error due to different lengths of the two chains to be superimposed"""
    def __init__(self, chain1, chain2):
        self.chain1=chain1
        self.chain2=chain2
    def __str__(self):
        return "The chains %s and %s can't be superimposed as they have different length." % (self.chain1.get_id(), self.chain2.get_id())

def checkCommands(commands):
    """Check the mode of operation of ComplexBuilder: if '-gui' has been defined
        it must be a unique argument, otherwise use the rest of commandline arguments"""
    inputs=[]
    data=[]
    if commands.gui: #enter files through GUI
        import complexconstructor.interface as interface
        if len(sys.argv)==2:
            inputList=interface.initGui()
            inputs = checkInputs(inputList[0], inputList[1]) #FASTA file, PDB dir
            data.append(inputs) # list of inputs : FASTA + PDB dir, data[0]
            data.append(inputList[3])  # stoichiometry None/Filename, data [1]
            checkOutput(inputList[4], data) #output file data[2]
            if  inputList[2]: # verbose T
                logProgress.logStart(inputs,data)
            data.append(inputList[2]) # verbose T/F data[3]
        else:
            raise ValueError("""
                If you want to use the graphical interface to
                introduce the complex data, the only commandline
                argument possible is '-gui'.\n
                Type -h for more information of the required
                format.""")
    else: #enter files through command line
        inputs = checkInputs(commands.infasta, commands.inpdb)
        data.append(inputs)
        checkSt(commands.stoich,data)
        checkOutput(commands.outfile, data)
        if commands.verbose: # if verbose is ON write progress in "ComplexBuilder.log"
            logProgress.logStart(inputs,data)
        data.append(commands.verbose) # verbose T/F data [3]

    return data

def checkSt(stName, inputsListed):
    """Check the stoichiometry file existence entered by the user and
    append it to the list of input values, otherwise add None"""
    if stName:
        if isfile(stName) and stName.endswith('.txt'):
            inputsListed.append(stName)
        else:
            raise ValueError("""
                You should introduce the name of an existing .txt
                file with the sctoichiometry data after '-st'.\n
                Type -h for more information of the required
                format.""")
    else:
        inputsListed.append(None)
    return None


def checkOutput(outputName, inputsListed):
    """Checks the output name existence defined by the user and
    append it to the list of input values, otherwise raise an error.
    Creates a new folder with the given name if it does not already
    exist, otherwise the results will be overwritten."""
    if outputName is not None:
        if not os.path.exists(outputName):
            os.mkdir(outputName)
            inputsListed.append(outputName)
        inputsListed.append(outputName)
    else:
        raise ValueError("""
                You should introduce the name of the directory
                where the resulting files will be stored after
                 '-o'.\n
                Type -h for more information of the required
                format.""")
    return None

def checkInputs(fastaFile, PDBDir):
    """Check if the FASTA file and PDB directory is introduced in
       the command line and returns all the files in a list"""

    if isfile(fastaFile)==False:
        raise ValueError("""
                You should introduce the name of the FASTA file
                containing the sequences of the elements of the
                complex you want to build after '-fa'.\n
                Type -h for more information of the required
                format.""")
    if PDBDir==None or os.path.exists(PDBDir)==False:
        raise ValueError("""
                You should introduce the name of the directory
                containing the PDB files for each interacting
                pair of the complex after '-pdb'.\n
                Type -h for more information of the required
                format.""")
    if isfile(fastaFile) and os.path.exists(PDBDir):
        if PDBDir.endswith("/"):
            string = ""
        else:
            string = "/"
        inputList=[]
        inputList = [PDBDir+string+f for f in os.listdir(PDBDir) if f.endswith(".pdb") and isfile(join(PDBDir, f))]
        if len(inputList)<1: # not PDB files inside PDBdir
            raise ValueError("""
                You should introduce the name of the directory
                containing the PDB files for each interacting
                pair of the complex after '-pdb'.\n
                Type -h for more information of the required
                format.""")
        if fastaFile.endswith(".fa") or fastaFile.endswith(".fasta"):
            inputList.append(fastaFile)
        else:
            raise ValueError("""
                You should introduce the name of the FASTA file
                containing the sequences of the elements of the
                complex you want to build after '-fa'.\n
                Type -h for more information of the required
                format.""")
        return inputList
    else:
        return None

def pdb_dna_to_sequence(chain):
    """This function receives a Chain object as input that is expected to be
    a DNA sequence, and returns the sequence as a string in single letter
    format."""
    seq = ""
    for res in chain:
        seq += res.get_resname()[-1]
    return seq

def data_extraction(pdb_files, fasta_file, verb, threshold = 0.90):
    """Takes as input a list of pdb files and a fasta file
    Returns a dictionary of dictionaries. The primary key is the model, the secondary
    key is the chain_id in said model and the value is the fasta_id of said chain.
    This function finds sequences by pairwise sequence alignment. If a pdb chain
    aligns with 0.95 (by default) identity with a fasta sequence, it is given the fasta identifier.
    If an aminoacid sequence is fragmented because of discontinuity or some other reason, the
    function puts it together."""
    big_dictionary = {}
    fasta_ids = []
    for pdb_file in pdb_files:
        model = PDBParser(QUIET = True).get_structure(pdb_file.split(".")[0], pdb_file)[0]
        big_dictionary[model] = {}
        for chain in model.get_chains():
            pdb_seqs = PPBuilder().build_peptides(chain)
            if len(pdb_seqs)>1:
                pp_seq = "".join(list([str(pp.get_sequence()) for pp in pdb_seqs]))
            elif len(pdb_seqs) == 1:
                pp_seq = pdb_seqs[0].get_sequence()
            else:
                pp_seq = pdb_dna_to_sequence(chain)
            for seq_record in SeqIO.parse(fasta_file,"fasta"):
                fasta_seq = seq_record.seq
                fasta_id = seq_record.id
                if fasta_id not in fasta_ids: fasta_ids.append(fasta_id)
                score = pairwise2.align.globalxx(fasta_seq,pp_seq, score_only = True)
                normalized_score = score/len(max([fasta_seq,pp_seq]))
                if (normalized_score >= threshold and abs(len(fasta_seq)-len(pp_seq))<=40):
                    big_dictionary[model][chain.get_id()] = fasta_id
                    break

    values = []
    tag=0
    for lista in [list(a.values()) for a in big_dictionary.values()]:
        values += lista
    for fasta_id in fasta_ids:
        if fasta_id not in set(values):
            tag += 1
            if verb: print("Fasta sequence with ID %s was not found between PDB files."%fasta_id,file = sys.stderr)
    if tag>4:
        raise SystemExit("Program stopped due to missing sequences.")
    return big_dictionary

def seq_dictionary(data):
    """Transforms the dictionary of dictionaries given by data_extraction
    into a dictionary with the fasta_ids as keys and as values, a list of
    interactions wich have:
     1: The model in which the sequence is found
     2: The corresponding chain in said model
     3: The other sequence (id) with which it is interacting"""
    sequence_dictionary = {}
    for k, v in data.items():
        for k2, v2 in v.items():
            sequence_dictionary.setdefault(v2, [])
            if list(k.get_chains()) in [list(a[0].get_chains()) for a in sequence_dictionary[v2]]: continue
            if list(v.values())[0]==list(v.values())[1]: other_id=list(v.values())[0]
            else: other_id = [seq_id for seq_id in v.values() if seq_id != v2][0]
            sequence_dictionary[v2].append((k,remove_heteroatoms(k[k2]),other_id))
    return sequence_dictionary

def stoichometry(file, information):
    """Takes as input a file with the stoichomety information and stores
    it in a dictionary format with sequences as keys and number of appearances
    in the complex as values.
    The sequence names should be the same as in the fasta file.
    The sequences found in the fasta and not in the stoichiometry file, are given
    stoichiometry 1."""
    fasta_ids = list(information.keys())
    dictionary = {}
    if file is not None:
        with open(file) as f:
            for line in f:
                line = line.strip().split(":")
                dictionary[line[0]] = int(line[1])
    for fasta_id in fasta_ids:
        if fasta_id not in list(dictionary.keys()):
            dictionary[fasta_id] = 1
    return dictionary

def remove_heteroatoms(chain):
    """Takes a chain as input and returns the same chain with heteroatoms removed."""
    chain_copy = copy.deepcopy(chain)
    heteroatoms = [atom for atom in chain.get_residues() if atom.id[0] != " "]
    for heteroatom in heteroatoms:
        chain_copy.detach_child(heteroatom.id)
    return chain_copy

def sequence_clashing(macrocomplex, third_chain):
    """Checks the number of clashes when adding a new chain to the macrocomplex.
    Takes as input a complex and a chain and uses pdb.NeighborSearch() in order to
    find how many CA atoms from the new chain are closer than 2 Angstroms to some
    CA atom in the complex. If it finds more than 20 atoms, it returns True, and so,
    the new chain will be discarted.
    """
    atoms_complex = [atom for atom in macrocomplex.get_atoms() if (atom.get_id() == 'CA' or atom.get_id() == 'P' or atom.get_id() == 'N1') ]
    atoms_chain = [atom for atom in third_chain.get_atoms() if (atom.get_id() == 'CA' or atom.get_id() == 'P' or atom.get_id() == 'N1')]
    neig_search = NeighborSearch(atoms_complex)
    n = 0
    for atom in atoms_chain:
         n+=len(list(neig_search.search(atom.get_coord(), 2.0, 'A')))
         if n >= 35:
             return True
    return False

def superimpositor(first_chain, same_chain, third_chain,macrocomplex):
    """ Adds new chain to the existing macrocomplex.
    Example: to add chain 'C' to the macrocomplex 'AB', when 'C' interacts with 'B', 'BC'
    This function takes as input 3 chain objects:
        1. 'first chain': the chain that we take as reference in
           order to do the superimposition -> 'B' from macrocomplex 'AB'
        2. 'same_chain': the chain that will be rotated because it interacts with the
            new chain (3) -> 'B' from interacting pair 'BC'
        3. 'third_chain': the new chain that we want to add to the macrocomplex -> 'C'
    Finally, it also takes the macrocomplex in order to add the third_chain-> 'AB'.
    It returns the new macrocomplex 'ABC' if 'C' does not clash with the previous
    structure 'AB'.
    """
    atom_list1 = Selection.unfold_entities(first_chain, 'A')
    atom_list2 = Selection.unfold_entities(same_chain, 'A')
    sup = Superimposer()
    chain_copy = copy.deepcopy(third_chain)

    if abs(len(atom_list1)-len(atom_list2))<30:
        if len(atom_list1)-len(atom_list2)>0:
            atom_list1=atom_list1[0:len(atom_list2)]
        elif len(atom_list1)-len(atom_list2)<0:
            atom_list2=atom_list2[0:len(atom_list1)]
    elif abs(len(atom_list1)-len(atom_list2))>=30:
        raise different_length_error(first_chain,same_chain)

    sup.set_atoms(atom_list1, atom_list2)
    sup.apply(chain_copy)
    sup.apply(third_chain)
    if sequence_clashing(macrocomplex,chain_copy):
        raise sequence_clashing_error(third_chain)
    else:
        N = 0
        while chain_copy.get_id() in [a.get_id() for a in macrocomplex.get_chains()]:
            try:
                chain_copy.id = alphabet[N]
            except ValueError:
                pass
            N += 1
        macrocomplex.add(chain_copy)
    return macrocomplex

def write_pdb(structure,path):
    """Writes a pdb file based of the specified structure in the given path."""
    io = PDBIO()
    io.set_structure(structure)
    io.save(path)

def constructor(information,stoich, verb):
    """Takes as input the dictionary created by the seq_dictionary function, the dictionary
    of stoichiometry created by the stoichometry function and the verbose parameter.
    Using the information from both dictionaries and the superimpositor function, it 
    builds the macrocomplex. 
    It returns the macrocomplex object after achieving the given stoichiometry.  
    """
    chains_in_complex={}

    #Choose a model to start iterating
    for seq, interactions in information.items():
        if len(interactions)>1 and (interactions[0][2]!=interactions[1][2] or stoich[interactions[0][2]]>1): break
    rand_interaction = information[seq][0]
    start_model = rand_interaction[0]
    first_chain = rand_interaction[1]
    chain1=copy.deepcopy(seq)

    #Store its chains in the complex dictionary and used chains list
    chains_in_complex[seq] = [first_chain]
    chains_in_complex[rand_interaction[2]] = [chain for chain in start_model.get_chains() if chain.get_id() != first_chain.get_id()]
    if verb: print("Chain %s has been correctly added. \nChain %s has been correctly added. " %(seq, rand_interaction[2]), file=sys.stderr)
    chains_used = [seq]
    #Make a copy to modify
    start_model_copy = copy.deepcopy(start_model)

    #Extend this model as much as possible from the first chain
    for interaction in information[seq]:
        if interaction[0] is start_model: continue
        if (interaction[2] in chains_in_complex) and (len(chains_in_complex[interaction[2]])>=stoich[interaction[2]]): continue
        second_model = interaction[0]
        same_chain = interaction[1]
        other_id = interaction[2]
        third_chain = [chain for chain in second_model.get_chains() if chain.get_id() != same_chain.get_id()][0]
        try:
            complex_out=superimpositor(first_chain, same_chain, third_chain, start_model_copy)
        except (sequence_clashing_error, different_length_error)  as error:
            if verb: print(error, file = sys.stderr)
        else:
            if verb: print("Chain %s has been correctly added." %other_id, file=sys.stderr)
            chains_in_complex.setdefault(other_id,[])
            chains_in_complex[other_id].append(third_chain)

    #From the resulting model, keep adding chains, until the model has as many chains as specified in the stoichiometry

    while len(list(complex_out.get_chains())) < sum(stoich.values()):
        #Get a sequence that is in the complex but is yet to be used as a core for the extension of the complex
        try:
            seq = [chain for chain in chains_in_complex if chain not in chains_used][0]
        #If it cannot find any, break the loop even if the complex is incomplete
        except IndexError:
            break

    #For each chain that is in the complex with this identifier
        for first_chain in chains_in_complex[seq]:
            #Add it to the used chains list
            chains_used.append(seq)
            #Inspect every interaction in this sequence identifier
            for interaction in information[seq]:
                #If it appears in the complex the same number of times times than it is specified in the stoichiometry, skip this interaction
                if (interaction[2] in chains_in_complex) and (len(chains_in_complex[interaction[2]])>=stoich[interaction[2]]): continue
                #Store the variables as in the first segment
                second_model = interaction[0]
                same_chain = interaction[1]
                other_id = interaction[2]
                #Get the other sequence
                third_chain = [a for a in second_model.get_chains() if a.get_id() != same_chain.get_id()][0]
                #Try adding it to the complex
                try:
                    complex_out=superimpositor(first_chain, same_chain, third_chain, complex_out)
                #If it does not work, do nothing
                except (sequence_clashing_error,different_length_error)  as error:
                    if verb: print(error, file = sys.stderr)
                #If it works, add it to the chains_in_complex dictionary
                else:
                    if verb: print("Chain %s has been correctly added." %other_id, file=sys.stderr)
                    chains_in_complex.setdefault(other_id,[])
                    chains_in_complex[other_id].append(third_chain)

    #We check if the stoichiometry is correct for all the chains
    chains_incomplete=[chain for chain in information if len(chains_in_complex[chain])<stoich[chain] ]
    for chain1 in chains_incomplete:
        found=False
        chains_interacting=[interaction[2] for interaction in information[chain1]]
        for chain in chains_interacting:
            if found:
                break
            interactions_with_chain1= [interaction for interaction in information[chain] if interaction[2]==chain1]
            for interaction in interactions_with_chain1:
                if found:
                    break
                for first_chain in chains_in_complex[chain]:
                    second_model = interaction[0]
                    same_chain = interaction[1]
                    third_chain = [a for a in second_model.get_chains() if a.get_id() != same_chain.get_id()][0]
                    try:
                        complex_out=superimpositor(first_chain, same_chain, third_chain, complex_out)
                    except (sequence_clashing_error,different_length_error)  as error:
                        if verb: print(error, file = sys.stderr)
                    else:
                        if verb: print("Chain %s has been correctly added." %chain1, file=sys.stderr)
                        chains_in_complex.setdefault(chain1,[])
                        chains_in_complex[chain1].append(third_chain)
                        if len(chains_in_complex[chain1])==stoich[chain1]:
                            found=True
                            break

    if verb:
        print("\nThe resulting complex has a total of %s chains. Its stoichiometry is the following: "%len(complex_out),file = sys.stderr)
        for id_, chains in chains_in_complex.items():
            print("\t%s : %s"%(id_, len(chains)), file = sys.stderr)

    return complex_out



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='complexconstructor',
      version='0.1.6',
      description='Generate macrocomplexes superimposing \
      paired interacting elements.',
      data_files = [("", ['LICENSE'])],
      license = "MIT",
      scripts=['bin/cconstruct'],
      long_description=long_description,
      long_description_content_type="text/markdown",
      keywords='macrocomplex bioinformatics structural pdb protein',
      url='https://github.com/Argonvi/SBI-PYT_Project',
      author='Paula Gomis Rosa, Arturo González Vilanova, Marta López Balastegui',
      author_email='marta.lopez01@estudiant.upf.edu, paula.gomis01@estudiant.upf.edu, arturo.gonzalez01@estudiant.upf.edu',
      packages=find_packages(),
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
        ],
      install_requires=['biopython'],
      include_package_data=True)
